"""
This module contains the class for aggregating datasets.
"""
from hyfi.utils.logging import LOGGING

logger = LOGGING.getLogger(__name__)


class DSAggregate:
    def __init__(self):
        pass
